package database

import (
	"crypto/sha256"
	"database/sql"
)

type ResourceHashIdentifier [sha256.Size]byte

func (h ResourceHashIdentifier) ToBytes() []byte {
	return h[:]
}

type Resource struct {
	Type
	id   int64
	hash ResourceHashIdentifier
	size int64
	path []byte //can be nil if not available locally
	mime string
}

func InsertResourceToDatabase(db *Database, hash ResourceHashIdentifier, size int64, path []byte, mime string) *Resource {
	rows, err := db.Query("INSERT INTO resources (hash, size, path, mime) VALUES ($1, $2, $3, $4) RETURNING id;", hash.ToBytes(), size, path, mime)
	if err != nil {
		return nil
	}
	rows.Next()
	defer rows.Close()
	var id int64
	err = rows.Scan(&id)
	if err != nil {
		return nil
	}
	return GetResourceFromDatabase(db, id)
}

func GetResourceFromDatabase(db *Database, id int64) *Resource {
	rows, err := db.Query("SELECT id, hash, size, path, mime FROM resources WHERE id = $1;", id)
	if err != nil {
		return nil
	}
	defer rows.Close()
	rows.Next()
	r, err := GetResourceFromRow(db, rows)
	if err != nil {
		return nil
	}
	return r
}

func GetResourceFromDatabaseByPath(db *Database, pathEntry string) *Resource {
	rows, err := db.Query("SELECT id, hash, size, path, mime FROM resources WHERE path = $1;", pathEntry)
	if err != nil {
		return nil
	}
	defer rows.Close()
	rows.Next()
	r, err := GetResourceFromRow(db, rows)
	if err != nil {
		return nil
	}
	return r
}

func GetResourceFromDatabaseByHash(db *Database, hash ResourceHashIdentifier) *Resource {
	rows, err := db.Query("SELECT id, hash, size, path, mime FROM resources WHERE hash = $1;", hash.ToBytes())
	if err != nil {
		return nil
	}
	defer rows.Close()
	rows.Next()
	r, err := GetResourceFromRow(db, rows)
	if err != nil {
		return nil
	}
	return r
}

func GetResourceFromRow(db *Database, row *sql.Rows) (*Resource, error) {
	r := &Resource{
		Type: Type{
			db: db,
		},
	}
	var hash []byte
	err := row.Scan(&r.id, &hash, &r.size, &r.path, &r.mime)
	if err != nil {
		return nil, err
	}
	copy(r.hash[:], hash)
	return r, nil
}

func (r *Resource) GetId() int64 {
	return r.id
}

func (r *Resource) GetHash() ResourceHashIdentifier {
	return r.hash
}

func (r *Resource) GetSize() int64 {
	return r.size
}

func (r *Resource) SetSize(size int64) {
	r.size = size
	r.db.Exec("UPDATE resources SET size = $2 WHERE id = $1;", r.id, r.size)
}

func (r *Resource) GetPath() []byte {
	return r.path
}

func (r *Resource) SetPath(path []byte) {
	r.path = path
	r.db.Exec("UPDATE resources SET mime = $2 WHERE id = $1;", r.id, r.path)
}

func (r *Resource) GetMime() string {
	return r.mime
}

func (r *Resource) SetMime(mime string) {
	r.mime = mime
	r.db.Exec("UPDATE resources SET mime = $2 WHERE id = $1;", r.id, r.mime)
}

func (r *Resource) GetReleases() (releases []*Release) {
	rows, err := r.db.Query("SELECT id, identifiers, metadata FROM releases WHERE id IN(SELECT release FROM resource_releases WHERE resource = $1);", r.id)
	if err != nil {
		return nil
	}
	defer rows.Close()
	for rows.Next() {
		release, err := GetReleaseFromRow(r.db, rows)
		if err != nil {
			break
		}
		releases = append(releases, release)
	}
	return
}

func (r *Resource) AddRelease(release *Release) {
	r.db.Exec("INSERT INTO resource_releases (resource, release) VALUES ($1, $2);", r.GetId(), release.GetId())
}
